from typing import Any, Callable, Iterable, Optional, Tuple, Union
from dataclasses import dataclass, field, asdict

try:
    import sqlite3
except ImportError:
    print(
        "I cannot import 'sqlite3'!"
    )
    exit(1)

class BaseConnection: # Connection Class
    """
    Connection class to create sqlite3 connection.

    ### Example:
    ```
    from eSQL import BaseConnection

    conn = BaseConnection('database.db') 
    # auto close database or can use conn.close()

    cursor = conn.cursor() # get connection cursor.
    # ...
    ```
    """

    conn: sqlite3.Connection = None
    cur: sqlite3.Cursor = None

    def __init__(self, *args, **kwargs) -> None:
        """
        open `SQLite3.Connection` object.
        """
        self.isolation_level = kwargs.get("isolation_level", None)

        self.conn = sqlite3.connect(*args, **kwargs, isolation_level=self.isolation_level)
        self.conn.row_factory = lambda c, r: dict(zip([col[0] for col in c.description], r))

        self.cur = self.conn.cursor()
    
    @property
    def is_autocommit(self) -> bool:
        return self.isolation_level is None

    @property
    def total_changes(self) -> Any:
        return self.conn.total_changes
    
    def commit(self) -> None:
        return self.conn.commit()
    
    def rollback(self, *args, **kwargs) -> None:
        return self.conn.rollback(*args, **kwargs)

    def load_extension(self, path: str) -> None:
        return self.conn.load_extension(path)
    
    def cursor(self) -> sqlite3.Cursor:
        """
        Return connection cursor.
        """
        return self.cur
    
    def new_cursor(self, cursorClass: Optional[type] = ...) -> sqlite3.Cursor:
        """
        Open new connection cursor object and return it.
        """
        return self.conn.cursor(cursorClass=cursorClass)
    
    def create_aggregate(self, name: str, n_arg: int, aggregate_class: Callable[[], Any]) -> None:
        self.conn.create_aggregate(name, n_arg, aggregate_class)
    
    def create_collation(self, __name: str, __callback: Any) -> None:
        self.conn.create_collation(__name, __callback)
    
    def create_function(self, name: str, narg: int, func: Any, *, deterministic: bool = ...) -> None:
        self.conn.create_function(name, narg, func, deterministic=deterministic)

    def backup(self, target: Union['BaseConnection', sqlite3.Connection], *, pages: int = ..., progress: Callable[[int, int, int], Union[object, None]] = ..., name: str = ..., sleep: float = ...) -> None:
        self.conn.backup(target.conn if (isinstance(target, BaseConnection)) else target, pages=pages, progress=progress, name=name, sleep=sleep)
    
    def foreign_key(self, enable: bool) -> None:
        self.cur.execute("PRAGMA foreign_key = %s" % ("OFF", "ON")[enable])

    def close(self) -> None:
        self.conn.close()
        self.conn = None
        self.cur = None
    
    def __del__(self) -> None:
        if not (self.conn is None):
            self.conn.close()

class _defaultKwargs(dict):
    sqlite_attrs = {
        "primary_key":   " PRIMARY KEY",
        "unique":        " UNIQUE",
        "default":       " DEFAULT {__item__}",
        "not_null":          " NOT NULL",
        "auto_increment":" AUTOINCREMENT",
        "max_length":    "({__item__})",
        "help":          " /* {__item__} */"
    }

    def __getitem__(self, k: Any) -> Any:
        value = self.get(k, None)
        if (value is None) or (value == False):
            return ""
        else:
            return self.sqlite_attrs.get(k, value).format(__item__=value)

    def __missing__(self, _) -> str:
        return ""

@dataclass
class _TypesParent:
    name: str = field(init=False, repr=False)
    _type: type = field(init=False, repr=False)
    help: str = None

# --- Integer Fields ---
@dataclass
class _IntegerParent(_TypesParent):
    _type = int

    primary_key: bool = False
    unique: bool = False
    default: int = None
    not_null: bool = False
    auto_increment: bool = False

    def to_statement(self) -> str:
        kwargs = asdict(self)
        return "{name}{primary_key}{unique}{not_null}{auto_increment}{default}{help}".format_map(_defaultKwargs(kwargs))

class SmallIntegerField(_IntegerParent):
    name = "SMALLINT"

class IntegerField(_IntegerParent):
    name = "INTEGER"

class MediumIntegerField(_IntegerParent):
    name = "MEDIUMINT"

class BigIntegerField(_IntegerParent):
    name = "BIGINT"

class UnsignedBigIntegerField(_IntegerParent):
    name = "UNSIGNED BIG INT"

# --- Character Fields ---
@dataclass
class _CharacterParent(_TypesParent):
    _type = str

    _max: int = field(init=False, default=None)
    max_length: int = None
    primary_key: bool = False
    unique: bool = False
    default: str = None
    not_null: bool = False
    
    def to_statement(self) -> str:
        kwargs = asdict(self)

        if kwargs["max_length"] is None:
            raise TypeError(
                "%s: max_length parameter needed." % (self.__class__.__name__)
            )
        
        if self._max:
            if (kwargs["max_length"] > self._max) or (kwargs["max_length"] < 0):
                raise ValueError(
                    "%s: max_length can be between 0 and %d." % (self.__class__.__name__, self._max)
                )
        
        return "{name}{max_length}{primary_key}{unique}{not_null}{default}{help}".format_map(_defaultKwargs(kwargs))

class CharField(_CharacterParent):
    _max = 255
    name = "VARCHAR"

class NCharField(_CharacterParent):
    name = "NCHAR"

# --- Text Field ---
@dataclass
class TextField(_TypesParent):
    name = "TEXT"
    _type = str

    primary_key: bool = False
    unique: bool = False
    default: str = None
    not_null: bool = False
    
    def to_statement(self) -> str:
        kwargs = asdict(self)
        return "{name}{primary_key}{unique}{not_null}{default}{help}".format_map(_defaultKwargs(kwargs))

# --- Fixed-Point Fields ---
@dataclass
class _FixedPointParent(_TypesParent):
    _type = float

    primary_key: bool = False
    unique: bool = False
    default: float = None
    not_null: bool = False
    auto_increment: bool = False

    def to_statement(self) -> str:
        kwargs = asdict(self)
        return "{name}{primary_key}{unique}{not_null}{auto_increment}{default}{help}".format_map(_defaultKwargs(kwargs))

class RealField(_FixedPointParent):
    name = "REAL"

class FloatField(_FixedPointParent):
    name = "FLOAT"

class DoubleField(_FixedPointParent):
    name = "DOUBLE"

# --- Decimal Field ---
@dataclass
class DecimalField(_TypesParent):
    name = "DECIMAL"
    _type = Any

    max_length: Union[int, Tuple[int, int]] = None
    primary_key: bool = False
    unique: bool = False
    default: Any = None
    not_null: bool = False

    def to_statement(self) -> str:
        kwargs = asdict(self)
        if kwargs.get("max_length", False):
            Var = kwargs["max_length"]
            kwargs["max_length"] = ",".join(str(i) for i in tuple(Var)) if (isinstance(Var, Iterable)) else str(Var)
    
        return "{name}{max_length}{primary_key}{unique}{not_null}{auto_increment}{default}{help}".format_map(_defaultKwargs(kwargs))

# --- Boolean Field ---
@dataclass
class BooleanField(_TypesParent):
    name = "BOOLEAN"
    _type = bool

    primary_key: bool = False
    unique: bool = False
    default: Union[int, bool] = None
    not_null: bool = False

    def to_statement(self) -> str:
        kwargs = asdict(self)
        return "{name}{primary_key}{unique}{not_null}{auto_increment}{default}{help}".format_map(_defaultKwargs(kwargs))
